//++
//display.c - MP3 LCD/VFD Display functions
//
// Copyright (C) 2005 by Spare Time Gizmos.  All rights reserved.
//
// This file is part of the Spare Time Gizmos' MP3 Player firmware.
//
// This firmware is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA.
//
//DESCRIPTION:
//   This module provides a simple interface to a generic display module, which
// may be either a Hitachi 44780 style generic text LCD, or a parallel ASCII
// vacuum flourescent display.  These displays maintain a cursor and current
// cursor location, but in general they (especially the LCDs) do not implement
// any terminal emulations, not even basic things like carriage return, line
// feed, tab, etc.  Thus you may call DisplayChar() several times in a row and
// assume that the characters are displayed right to left, but if you want to
// move to the next line you need to use one of the cursor position functions
// to move the cursor there.  A line feed or CRLF won't do it.
//
//   Needless to say, the LCD doesn't implement scrolling either.  Your only
// option is to erase the entire display and repaint.  Also note that this
// code does not enforce any display size limits, so it's up to the application
// code not to exceed the display width and height.  The DISPLAY_WIDTH and
// DISPLAY_HEIGHT pre-processor constants are just the thing you need for this.
//
//REVISION HISTORY:
// dd-mmm-yy    who     description
// 18-May-05	RLA	New file.
// 13-Jul-05	RLA	Add a small delay to WaitBusy() for VFDs...
// 20-Jul-05	RLA	Add the copyright sign to the LCD's CGRAM ...
// 12-Oct-05	RLA	For VFDs, home cursor THEN clear screen
// 15-Oct-05	RLA	Remove DISPLAY_TYPE and add LCD/VFD conditionals
// 20-Oct-05	RLA	Fix a timing bug in the VFD DisplayWrite()
//			
//--

// Include files...
#include <stdio.h>		// needed so DBGOUT(()) can find printf!
#include <stdarg.h>		// va_alist(), va_end(), etc...
#include "standard.h"		// standard types - BYTE, WORD, BOOL, etc
#include "reg66x.h"		// declarations for Philips 89C66x processors
#include "player.h"		// project wide (hardware configuration) declarations
#include "debug.h"		// debugging stuff and ResetWDT()
#include "post.h"		// POST codes for Fail()
#include "timer.h"		// DELAYUS() macro and DelayMS() function
#include "display.h"		// declarations for this module

//   This array of eight bytes defines a custom character (a copyright symbol)
// for the Hitachi 44780 LCD controller.  The VFD character set already contains
// a copyright sign, so this is unnecessary there...
#ifdef LCD
PRIVATE CODE const BYTE m_abCopyrightSymbol[8] = {
  0x0E, 0x11, 0x17, 0x19, 0x17, 0x11, 0x0E, 0x00
};
#endif

// Local storage...
PRIVATE XDATA char m_acDisplayBuffer[DISPLAY_WIDTH+1];


//++
//   This routine will write one byte to an LCD display.  For an LCD display,
// bRS selects the control register (bRS=1) or the data register (bRS=0).
// bData gives the actual 8 bit data to be sent to the display.  Note that
// this routine does NOT check the busy flag before sending the data
// - that's the responsibility of the caller!
//
//   VFD displays are similar, however in this case there is only a data port
// and no register select flag is used.  Also, VFD displays use the opposite
// polarity for the ENABLE signal (active low for VFDs rather than active high
// for LCDs)...
//--
#ifdef LCD
PRIVATE void DisplayWrite (BYTE bRS, BYTE bData)
{
  DISPLAY_RW = 0;  DISPLAY_RS = (bRS != 0);  DISPLAY_DATA = bData;
  DISPLAY_ENABLE = 1;  DELAYUS(2);  DISPLAY_ENABLE = 0;
}
#endif
#ifdef VFD
PRIVATE void DisplayWrite (BYTE bData)
{
  INT_OFF;
  DISPLAY_RW = 1;  DISPLAY_ENABLE = 0;  DISPLAY_DATA = bData;
  DISPLAY_RW = 0;  DISPLAY_RW = 1;
  DISPLAY_ENABLE = 1;  DISPLAY_RW = 0;
  INT_ON;
  //DISPLAY_RW = 0;  DISPLAY_DATA = bData;
  //DISPLAY_ENABLE = 0;  DELAYUS(2);  DISPLAY_ENABLE = 1;
}
#endif


//++
//   This routine will monitor the display's BUSY flag and (contrary to
// what the name might lead you to believe :-) it will wait for the display
// to NOT be busy.
//
//   LCD displays implement the BUSY flag as the MSB of the status register,
// which must be explicitly read by setting the RW bit, clearing the register
// select, and then enabling the display.
//
//   VFD displays have a completely separate BUSY flag which is connected to
// it's own port pin.  We can simply test this pin without any other effort.
//
//   Note that both versions implement a timeout which will eventually POST
// if the BUSY flag never clears...
//--
PRIVATE void WaitBusy (void)
{
  WORD wTimeout = 0;
#ifdef LCD
  //  Set the RW bit and clear RS to read the status register, then set the data
  // port to all 1s so that it can be used as an input.
  DISPLAY_RW = 1;  DISPLAY_RS = 0;  DISPLAY_DATA = 0xFF;  DISPLAY_ENABLE = 1;
  while (TRUE) {
    DELAYUS(1);
    if ((DISPLAY_DATA & 0x80) == 0) break;
    if (++wTimeout == 0) POST(PC_DISPLAY, FALSE);
  }
  // Disable the display and return...
  DISPLAY_ENABLE = 0;
#endif
#ifdef VFD
  // VFDs have an explicit status line that makes things much easier...
  while (TRUE) {
    //   I had a lot of trouble with glitches on the DISPLAY_BUSY signal,
    // especially with a faster 89C66x clock (the faster CPU clock doesn't
    // cause the glitches - it just makes it easier for the software to find
    // them!)  The only fix is to sample DISPLAY_BUSY and, if we find it
    // cleared, wait a few microseconds and then sample again to be sure we
    // get the same result...
    if (!DISPLAY_BUSY) {
      DELAYUS(2);  if (!DISPLAY_BUSY) return;
    }
    if (++wTimeout == 0) POST(PC_DISPLAY, FALSE);
  }    
#endif
}


//++
//   This routine will intitialize the display and clear the screen.  Normally
// it's called only once, during startup, but it could be called again if
// necessary.  Note that it contains several programmed delays to wait for the
// LCD to initialize, so it isn't especially fast!
//
//   Although we never explicitly read back from the display, we do explicitly
// check the display's BUSY flag for both LCD and VFD displays.  If no display
// is connected (or if it's broken somehow) then we will eventually time out and
// POST() via the WaitBusy() routine...
//--
PUBLIC void InitializeDisplay (void)
{
#ifdef LCD
  BYTE i;
  //   LCD displays are notorious for being difficult to initialize, particularly
  // when the power supply startup is noisy or slow and glitches the display's
  // microprocessor.  The solution is to follow the data book initialization
  // procedure explicity, which is exactly what we do here (in this case it's the
  // Optrex data book we're following)...
  //
  //   Also note that LCDs have no hardware reset, so the DISPLAY_RESET signal
  // does nothing.  And finally, for LCD dispays the ENABLE signal is active
  // high, and so we initialize it to low...
  DISPLAY_ENABLE = 0;  DelayMS(15);
  //   At this point it's not safe to depend on the display's BUSY flag, and
  // we must make do with programmed delays instead...
  DisplayWrite(CONTROL_REGISTER, LCD_FUNCTION_SET|LCD_8BIT);		     	DelayMS(5);
  DisplayWrite(CONTROL_REGISTER, LCD_FUNCTION_SET|LCD_8BIT);		     	DELAYUS(160);
  // Now we can start using the BUSY flag...
  DisplayWrite(CONTROL_REGISTER, LCD_FUNCTION_SET|LCD_8BIT);		     	WaitBusy();
#if (DISPLAY_HEIGHT == 1)
  DisplayWrite(CONTROL_REGISTER, LCD_FUNCTION_SET|LCD_8BIT);		   	WaitBusy();
#else
  DisplayWrite(CONTROL_REGISTER, LCD_FUNCTION_SET|LCD_8BIT|LCD_TWO_LINES);   	WaitBusy();
#endif
  DisplayWrite(CONTROL_REGISTER, LCD_DISPLAY_CONTROL|LCD_DISPLAY_OFF);	     	WaitBusy();
  DisplayWrite(CONTROL_REGISTER, LCD_CLEAR_DISPLAY);			     	WaitBusy();
  DisplayWrite(CONTROL_REGISTER, LCD_CURSOR_MODE_SET);			     	WaitBusy();
  DisplayWrite(CONTROL_REGISTER, LCD_ENTRY_MODE_SET|LCD_INCREMENT_CURSOR);	WaitBusy();
  DisplayWrite(CONTROL_REGISTER, LCD_DISPLAY_CONTROL|LCD_DISPLAY_ON);		WaitBusy();
  // Load the character generate RAM on the 44780...
  DisplayWrite(CONTROL_REGISTER, LCD_LOAD_CGRAM);  WaitBusy();
  for (i = 0;  i < sizeof(m_abCopyrightSymbol);  ++i) {
    //   Note that the actual character generator data is written to the LCD's
    // _data_ port, not the control port!
    DisplayWrite(DATA_REGISTER, m_abCopyrightSymbol[i]);  WaitBusy();
  }  
#endif
#ifdef VFD
  //   VFD displays have a hardware RESET pin, and also have no especially nasty
  // startup conditions, so our life is much easier.  One gotcha, however, is that
  // VFD displays use an active low ENABLE signal, so we initialize it to HIGH.
  DISPLAY_ENABLE = 1; 
#ifdef PCBV1
  DISPLAY_RESET = 0;  DelayMS(20);  DISPLAY_RESET = 1;
#endif
  DelayMS(20);
  WaitBusy();  DisplayWrite(VFD_ESCAPE);
  WaitBusy();  DisplayWrite(VFD_FLICKERLESS);
  WaitBusy();
#endif
  DBGOUT(("Display initialized ...\n"));
}


//++
// Erase (clear) the entire display and move the cursor to (0,0).
//--
PUBLIC void ClearDisplay(void)
{
#ifdef LCD
  WaitBusy();  DisplayWrite(CONTROL_REGISTER, LCD_CLEAR_DISPLAY);
  WaitBusy();  DisplayWrite(CONTROL_REGISTER, LCD_CURSOR_HOME);  
#endif
#ifdef VFD
  WaitBusy();  DisplayWrite(VFD_CURSOR_HOME);
  WaitBusy();  DisplayWrite(VFD_CLEAR_DISPLAY);
  WaitBusy();
#endif
  //DBGOUT(("Clear display\n"));
}


//++
//   Move the display's cursor (which isn't necessarily visible!) to the 
// selected row and column, the upper left being (0,0).  The behaviour is
// undefined (although generally it's a NOP) if bRow > DISPLAY_HEIGHT-1 or
// bCol > DISPLAY_WIDTH-1.
//--
PUBLIC void DisplayAt(BYTE bRow, BYTE bCol)
{
  BYTE bPos;
  if ((bRow >= DISPLAY_HEIGHT) || (bCol >= DISPLAY_WIDTH)) {
    DBGOUT(("???Bad cursor position bRow=%bd, bCol=%bd\n", bRow, bCol));
    return;
  }
#ifdef LCD
  //   Cursor addressing on the LCD is really ugly!  For 2x40 displays, the
  // cursor position ranges from 0x00..0x27 (39 decimal) for the first line
  // and 0x40..0x67 for the second line.  Four 4x20 displays, however, the
  // first line is 0x00..0x13, the second line is 0x40..0x53, the third line
  // is 0x14..0x27 (i.e. it's the second half of the first line!) and the
  // fourth line is 0x54..0x67.  Uggh!
#if (DISPLAY_HEIGHT == 2)
  bPos = (bRow == 0 ? 0x00 : 0x40) | bCol;
#elif (DISPLAY_HEIGHT == 4)
  switch (bRow) {
    case 0: bPos = bCol;			break;
    case 1: bPos = 0x40 | bCol;			break;
    case 2: bPos = DISPLAY_WIDTH+bCol;          break;
    case 3: bPos = 0x40 | (DISPLAY_WIDTH+bCol);	break;
  }
#else
  bPos = bCol;
#endif
  WaitBusy();  DisplayWrite(CONTROL_REGISTER, LCD_SET_DDRAM_ADDRESS | bPos);
#endif
#ifdef VFD
  //   For VFD displays it's much easier.  First, all display positions are
  // numbered sequentially (e.g. the first character of the first line is at
  // location 0, the first character of the second line is at DISPLAY_WIDTH,
  // DISPLAY_WIDTH*2 for the third line, etc).
  bPos = (bRow * DISPLAY_WIDTH) + bCol;
  WaitBusy();  DisplayWrite(VFD_ESCAPE);
  WaitBusy();  DisplayWrite(VFD_SET_CURSOR);
  WaitBusy();  DisplayWrite(bPos);
  WaitBusy();
#endif
  //DBGOUT(("Set Display Cursor Row=%bd, Column=%bd (bPos=0x%02bX)\n", bRow, bCol, bPos));
}

#ifdef UNUSED
//++
//   Turn the actual cursor display on (bShow=TRUE) or off (bShow=FALSE).  This
// is strictly a visual effect - text output, including DisplayAt(), is not
// affected.
//--
PUBLIC void DisplayCursor (BOOL bShow)
{
#ifdef LCD
  BYTE bCursor =  LCD_DISPLAY_CONTROL|LCD_DISPLAY_ON;
  if (bShow) bCursor |= LCD_CURSOR_ON|LCD_CURSOR_BLINK;
  WaitBusy();  DisplayWrite(CONTROL_REGISTER, bCursor);
#endif
#ifdef VFD
  WaitBusy();  DisplayWrite (bShow ? VFD_CURSOR_ON : VFD_CURSOR_OFF);
#endif
  //DBGOUT(("Display Cursor, bShow=%bd\n", bShow));
}
#endif

//++
//   Display a single character at the current cursor location and move the
// cursor left one position.  What happens when the cursor exceeds the display
// width is undefined (and in the case of most multi-line LCD displays, at
// least, can be quite creative :-)
//--
PUBLIC void DisplayChar (char bChar)
{
#ifdef LCD
  WaitBusy();  DisplayWrite(DATA_REGISTER, bChar);
#endif
#ifdef VFD
  WaitBusy();  DisplayWrite(bChar);  /*WaitBusy();*/
#endif
  //DBGOUT(("DisplayChar, bChar=%bd ('%c')\n", bChar, bChar));
}

#ifdef UNUSED
//++
//   This function is equivalent to a call to DisplayChar() with an implied
// call to DisplayCursor() first...
//--
PUBLIC void DisplayCharAt (BYTE bRow, BYTE bCol, char bChar)
{
  DisplayAt(bRow, bCol);  DisplayChar(bChar);
}
#endif

//++
//   This routine simply calls DisplayChar() multiple times.  Note that the
// string pointer parameter is deliberately an "un-spaced" declaration, so
// that it can accomodate argument strings in the CODE, DATA, IDATA or XDATA
// memory spaces..
//--
PUBLIC void DisplayString (char *pString)
{
  while (*pString != '\0') DisplayChar(*pString++);
  //DBGOUT(("DisplayString, pString=\"%s\"\n", pString));
}

//++
// Just like DisplayString(), but with a call to DisplayCursor() first...
//--
PUBLIC void DisplayStringAt (BYTE bRow, BYTE bCol, char *pString)
{
  DisplayAt(bRow, bCol);  DisplayString(pString);
}


#ifdef UNUSED
//++
//   This procedure outputs an sprintf() formatted string to the display 
// starting at the current cursor position.  It assumes that the string
// length never exceeds the DISPLAY_WIDTH - bad, bad, bad things will
// happen if you violate this rule!!
//--
PUBLIC void DisplayText (char *pFormat, ...)
{
  va_list arg_ptr;  WORD wCount;
  va_start (arg_ptr, pFormat);
  wCount = vsprintf(m_acDisplayBuffer, pFormat, arg_ptr);
  ASSERT((wCount <= DISPLAY_WIDTH+1), ("DisplayText: \"%s\" too long (%d chars)\n", pFormat, wCount));
  va_end (arg_ptr);
  DisplayString(m_acDisplayBuffer);
  //DBGOUT(("DisplayText, text=\"%s\"\n", m_acDisplayBuffer));
}
#endif

//++
//  The same as DisplayText() with a call to DisplayCursor() first...
//--
PUBLIC void DisplayTextAt (BYTE bRow, BYTE bCol, char *pFormat, ...)
{
  va_list arg_ptr;  WORD wCount;
  va_start (arg_ptr, pFormat);
  wCount = vsprintf(m_acDisplayBuffer, pFormat, arg_ptr);
  ASSERT((wCount <= DISPLAY_WIDTH+1), ("DisplayTextAt: \"%s\" too long (%d chars)\n", pFormat, wCount));
  va_end (arg_ptr);
  DisplayStringAt(bRow, bCol, m_acDisplayBuffer);
  //DBGOUT(("DisplayText, text=\"%s\"\n", m_acDisplayBuffer));
}


#if 0
PUBLIC void DisplayTest (void)
{
  BYTE x=1, y=2;
  ClearDisplay();
  DisplayStringAt(0,0,"ROW 1!!");
  printf("Press a key...");  _getkey();
  DisplayStringAt(1,1,"ROW 2!!");
  printf("Press a key...");  _getkey();
  DisplayStringAt(2,2,"ROW 3!!");
  printf("Press a key...");  _getkey();
  DisplayStringAt(3,3,"ROW 4!!");
  printf("Press a key...");  _getkey();

  ClearDisplay();
  DisplayCharAt(0,18,'X');  DisplayChar('Y');
  DisplayStringAt(1,0,"CURSOR ON!");  DisplayCursor(TRUE);
  printf("Press a key...");  _getkey();
  DisplayStringAt(2,0,"CURSOR OFF!");  DisplayCursor(FALSE);
  printf("Press a key...");  _getkey();

  ClearDisplay();
  DisplayStringAt(0,0,"DisplayText() Test");
  DisplayTextAt(1,4,"x=%bd", x);
  DisplayText(", y=%bd", y);
  DisplayTextAt(2,10,"x+y=%bd", (BYTE) (x+y));
  //BYTE b;
  //DisplayString("THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG\n");
  //DisplayString("test");
  //for (b = 1;  b < 32;  ++b)  DisplayWrite(1, 96+b);
}
#endif

