//++
//fat.c - MP3 Player Compact Flash Card functions
//
// Copyright (C) 2005 by Spare Time Gizmos.  All rights reserved.
//
// This file is part of the Spare Time Gizmos' MP3 Player firmware.
//
// This firmware is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 Temple
// Place, Suite 330, Boston, MA  02111-1307  USA.
//
//DESCRIPTION:
//   This module knows just enough about FAT file systems to a) mount the first
// partition on the CompactFlash card, b)list all the files in the root directory,
// and b) find and read the data in an individual file.  Other than that it's
// pretty limited.  It understands ONLY FAT16 file systems - neither FAT12 nor
// FAT32 are supported.  It also only knows how to read files; there's nothing to
// create files nor allocate free disk space for writing them!  This code does,
// however, have enough sanity checking to reject Compact Flash cards which don't
// meet our requirements.
//
//REFERENCES:
//	http://www.compuphase.com/mbr_fat.htm
//	http://support.microsoft.com/kb/q140418/
//	http://encyclopedia.thefreedictionary.com/FAT16
//	http://www.cse.scu.edu/~tschwarz/coen152_05/Lectures/FAT.html
//
//REVISION HISTORY:
// dd-mmm-yy    who     description
// 20-May-05	RLA	New file.
// 20-Jul-05	RLA	Change NextFATFile() and PreviousFATFile() so that they
//			 loop back when they reach the end/beginning of the root
//			 directory.  This way they just repeat the same sequence
//			 of files over and over again - just what we want an MP3
//			 player to do.
// 12-Oct-05	RLA	Change a lot of things that were parameters (e.g. volume
//			 label, FAT type, etc) into global static variables to
//			 make life easier for SDCC.
// 13-Oct-05	RLA	Semi-major changes to a) remove all statically allocated
//			 disk buffers - instead, buffers are pulled from the free
//			 pool as needed - and b) to allow the directory sector
//			 buffer to be freed while an MP3 is actually playing. This
//			 gives us an extra buffer for MP3 data, which makes it
//			 possible to run on an 89C664 MCU with only 2K RAM.
// 18-Oct-05	RLA	Add basic support for MBRs
// 19-Oct-05	RLA	Make the current FAT buffer statically allocated in XDATA.
//			 Since it's permanently allocated there's no real reason
//			 to use a disk buffer, and doing so wastes a BCB entry and
//			 lots of IDATA space that SDCC can't afford (it works great
//			 with C51, though!).
//			Move ID3 tag stuff to id3.c
//			A "return TRUE" was accidentally left out of StartDirScan()...
//			Add STACKSLOCS hack for SDCC (see slocs.h for more info)...
// 20-Oct-05	RLA	Don't crash if we find an actual LFN longer than MAX_LFN_LENGTH
//			Fix a little endian problem with GetFileLFN()
//--

// Include files...
#include <stdio.h>		// needed so DBGOUT(()) can find printf!
#include <string.h>		// strcpy(), etc...
#include "standard.h"		// standard types - BYTE, WORD, BOOL, etc
#include "cfcard.h"		// Compact Flash card I/O functions
#include "buffer.h"		// Buffer management functions
#include "fat.h"		// declarations for this module
#include "id3.h"		// ID3 v1.1 definitions


// Public members...
//   In the original C51 version, these items were usally automatic variables in another
// module (generally player.c) and passed as parameters to routines in this module.  SDCC,
// however, isn't all that swift at dealing with parameters and local variables, so they
// were converted to globals in the interest of simplifying the generated code.  Sadly, 
// the code quality suffers somewhat as a result, but I guess that's why Keil can charge
// the big bucks...
PUBLIC XDATA char g_szVolumeLabel[VOLUME_LABEL_LENGTH];	// GetVolumeData()
PUBLIC XDATA char g_szSystemType[SYSTEM_TYPE_LENGTH];	// InitializeFAT()
PUBLIC XDATA WORD g_wTotalMP3Files;			// GetVolumeData()
PUBLIC XDATA char g_szLongFileName[MAX_LFN_LENGTH];	// GetFileLFN()

// Volume data...
PRIVATE XDATA LONG  m_lVolFATOffset;	// offset (in sectors) of the first FAT
PRIVATE XDATA LONG  m_lVolRootOffset;	//  "  "   "    "  "   "   "  root directory
PRIVATE XDATA LONG  m_lVolDataOffset;	//  "  "   "    "  "   "   "  file data area
PRIVATE XDATA LONG  m_lVolFATSize;	// size of each FAT (in sectors)
PRIVATE XDATA LONG  m_lVolRootSize;	// size of the root directory (in sectors!)
PRIVATE XDATA BYTE  m_bVolClusterSize;	// sectors per cluster on this device
// Current directory sector status...
PRIVATE DATA  BYTE  m_bCurDirEntry;	// offset of the current directory entry
PRIVATE DATA  LONG  m_lCurDirSector;	// sector currently in directory buffer
//   The next member points to the buffer used to read directory sectors.  This can
// be, and often is, NULL, indicating that there is no buffer allocated.  Especially
// while we're playing, the directory buffer is released to free up more memory for
// MP3 file data.  Calling ReadDirSector() will allocate another buffer...
PRIVATE DATA  PIBCB m_pCurDirBCB;	// BCB of current directory sector
// Current FAT sector status...
PRIVATE IDATA LONG  m_lCurFATSector;	// LBN of the current sector (in m_wCurFAT)...
PRIVATE XDATA CLUSTER m_awCurFAT[CLUSTERS_PER_SECTOR];
// Current file status...
PRIVATE IDATA WORD  m_wCurFileCluster;	// current cluster we're reading
PRIVATE IDATA int   m_nRelFileSector;	// relative sector within the cluster
PRIVATE IDATA LONG  m_lCurFileSector;	// absolute sector (Cluster + RelSector)
PRIVATE IDATA LONG  m_lFileBytesLeft;	// bytes remaining before EOF
//   This is a temporary directory entry which NextMP3File() and PrevMP3File() use
// to cache the current file data while we're playing an MP3 file...
PRIVATE XDATA DIRECTORY_ENTRY m_CurDirEntCache;

//   This macro generates a pointer (of type PXDIRENT) to file #n in the current
// directory sector, m_pCurDirBCB.  The index n can safely range from 0 to
// DIRECTORY_ENTRIES_PER_SECTOR-1.  It's just a metter of caste ...
#define GetDirEntry(n)	( (PXDIRENT) (m_pCurDirBCB->pbBuffer + (n)*sizeof(DIRECTORY_ENTRY)) )
//   This macro is essentially the same, except that it always returns the
// current directory entry, as determined by m_bCurDirEntry...
#define GetCurDirEntry()  GetDirEntry(m_bCurDirEntry)


//++
//   This debugging routine will dump out the contents of a sector buffer in
// both hexadecimal and ASCII.  There's not much too it!
//--
#if 0
PRIVATE void DumpSector (PXBYTE pxBuffer)
{
  WORD wOffset, i;  char ch;
  for (wOffset = 0;  wOffset < IDE_SECTOR_SIZE;  wOffset += 16) {
    printf("%03X/ ", wOffset);
    for (i = 0;  i < 16;  ++i)
      printf(" %02bx", pxBuffer[wOffset+i]);
    printf("    ");
    for (i = 0;  i < 16;  ++i) {
     ch = pxBuffer[wOffset+i] & 0x7F;
     if ((ch < ' ') || (ch >= 0x7F)) ch = '.';
     printf("%c", ch);
    }
    printf("\n");
  }
}
#endif

//++
// Show one entry from the partition table in the MBR.  For debugging only!
//--
#ifdef SHOWPARTITION
PRIVATE void ShowPartition (BYTE bPart, MASTER_BOOT_RECORD XDATA *pxMBR)
{
  DBGOUT(("Partition #%bd - Flags = 0x%02bX, Type = 0x%02bX\n",
    bPart, pxMBR->aPartitions[bPart].bFlags, pxMBR->aPartitions[bPart].bType));
  DBGOUT(("\tStartCHS = 0x%02bX %02bX %02bX, EndCHS = 0x%02bX %02bX %02bX\n",
    pxMBR->aPartitions[bPart].abStartCHS[0], pxMBR->aPartitions[bPart].abStartCHS[1], 
    pxMBR->aPartitions[bPart].abStartCHS[2], pxMBR->aPartitions[bPart].abEndCHS[0],
    pxMBR->aPartitions[bPart].abEndCHS[1],  pxMBR->aPartitions[bPart].abEndCHS[2]));
  DBGOUT(("\tOffset (LBN) = %ld, Size (sectors) = %ld\n",
    LSWAP(pxMBR->aPartitions[bPart].lOffset), LSWAP(pxMBR->aPartitions[bPart].lSize)));
  DBGOUT(("Master boot record signature = 0x%04X\n", WSWAP(pxMBR->wSignature)));
}
#endif

//++
//   This debugging routine will print the contents of the boot sector.
// It's invaluable for figuring out what's wrong...
//--
#ifdef SHOWBOOTSECTOR
PRIVATE void ShowBootSector (BOOT_SECTOR XDATA *pxVBS)
{
  BPB XDATA *pxBPB = &(pxVBS->bpb);
  printf("Volume Boot Sector\n");
  printf("\tabJump = 0x%02bX %02bX %02bX \n", pxVBS->abJump[0], pxVBS->abJump[1], pxVBS->abJump[2]);
  printf("\tszOEMName[8] = \"%8s\"\n", pxVBS->szOEMName);
  printf("\twBootSignature = 0x%04X\n", WSWAP(pxVBS->wBootSignature));
  printf("BIOS Parameter Block\n");
  printf("\twBytesPerSector = %d\n", WSWAP(pxBPB->wBytesPerSector));
  printf("\tbSectorsPerCluster = %bd\n", pxBPB->bSectorsPerCluster);
  printf("\twReservedSectors = %d\n", WSWAP(pxBPB->wReservedSectors));
  printf("\tbNumberFATs = %bd\n", pxBPB->bNumberFATs);
  printf("\twRootEntries = %d\n", WSWAP(pxBPB->wRootEntries));
  printf("\twSmallSectors = %d\n", WSWAP(pxBPB->wSmallSectors));
  printf("\tbMediaDescriptor = 0x%02bX\n", pxBPB->bMediaDescriptor);
  printf("\twSectorsPerFAT = %d\n", WSWAP(pxBPB->wSectorsPerFAT));
  printf("\twSectorsPerTrack = %d\n", WSWAP(pxBPB->wSectorsPerTrack));
  printf("\twTotalHeads = %d\n", WSWAP(pxBPB->wTotalHeads));
  printf("\tlHiddenSectors = %ld\n", LSWAP(pxBPB->lHiddenSectors));
  printf("\tlLargeSectors = %ld\n", LSWAP(pxBPB->lLargeSectors));
  printf("\tbPhysicalDrive = %bd\n", pxBPB->bPhysicalDrive);
  printf("\tbCurrentHead = %bd\n", pxBPB->bCurrentHead);
  printf("\tbBootSignature = 0x%02bX\n", pxBPB->bBootSignature);
  printf("\tlVolumeID = 0x%08lX\n", LSWAP(pxBPB->lVolumeID));
  printf("\tszVolumeLabel[11] = \"%11.11s\"\n", &(pxBPB->szVolumeLabel));
  printf("\tszSystemID[8] = \"%8.8s\"\n", &(pxBPB->szSystemID));
}
#endif


//++
//   This routine will read sector zero from the flash card and attempt
// to determine if it contains a valid master boot record.  If an MBR
// is present, then this code uses the partition table in the MBR to
// find and return the first sector of the first partition on the drive.
// If no MBR is found, then it returns zero.
//
//   The value returned thus tells us where to go to look for the volume
// boot sector (VBS), which is the next step in reading the file system.
// Notice that it's no accident that zero is returned if no MBR is found,
// because in the case of flash cards that have been formatted without
// any MBR (i.e. they're formatted as if they were a big floppy disk) then
// sector zero is where we should look next for the VBS.
//
//   BTW, I know of no single, definitive, test to determine whether a volume
// has an MBR or not.  Unfortunately the 0xAA55 signature that's stored at the
// end of the MBR just happens to be the same signarture, and stored in the
// same place, that you'd find in a VBS.  Worse, most of the MBR is usually
// filled with zeros, and the partition table is the only part which holds any
// meaningful data.  That doesn't give us a lot to go on!
//--
PRIVATE LONG ReadMBR (void) STACKSLOCS
{
  PIBCB pBCB;  MASTER_BOOT_RECORD XDATA *pxMBR;  BYTE i;

  //   Allocate a buffer and read the master boot record.  Note that although
  // the VBS can move around, the MBR is ALWAYS in first sector on the disk!
  ALLOCATE_BUFFER(pBCB);
  if (pBCB == NULL) return 0L;
  //DBGOUT(("ReadMBR: reading master boot record from LBN 0\n"));
  if (!ReadSector(0L, pBCB->pbBuffer)) goto FAIL;
  pxMBR = (MASTER_BOOT_RECORD XDATA *) pBCB->pbBuffer;

  //  If the signature isn't valid, then this is neither a master boot record
  // nor a volume boot sector.  There isn't much we can do except give up!
  if (WSWAP(pxMBR->wSignature) != 0xAA55) goto FAIL;

  // Go thru all the partition table entries until we find one that's usable.
  for (i = 0;  i < MAX_PARTITIONS;  ++i) {
#ifdef SHOWPARTITION
    ShowPartition(i, pxMBR);
#endif
    //   Check the system type to see if this partition is a DOS FAT16 volume.
    // Unfortunately there are several types which are valid!
    if (   (pxMBR->aPartitions[i].bType != 0x04)   	// FAT16 for partitions <= 32Mb
        && (pxMBR->aPartitions[i].bType != 0x06)    	// FAT16 for partitions > 32Mb
        && (pxMBR->aPartitions[i].bType != 0x0E) ) 	// BIGDOS FAT16 partition
      continue;
    //   For the partition to be valid, the flags have to be either 0x00 (normal
    // partition) or 0x80 (a bootable partition).  The latter isn't likely, but
    // we might as well accept it if we find one...
    if (   (pxMBR->aPartitions[i].bFlags != 0x00)
        && (pxMBR->aPartitions[i].bFlags != 0x80) ) continue;
    //   If either of the LBA fields are zero (which normally means this disk uses
    // CHS addressing) then we can't do anything with this partition.  [BTW, notice
    // that these two tests should technically have a LSWAP() in them, but when we're
    // simply testing against zero the byte order makes no difference!]
    if (   (pxMBR->aPartitions[i].lOffset == 0L)
        || (pxMBR->aPartitions[i].lSize   == 0L) ) continue;
    //  The last test, which is the only real sanity test we do, is to verify
    // that the last sector in the partition (i.e. the partition offset plus
    // the partition size) is less than the physical size of the CF card.  It
    // doesn't prove the MBR is valid, but at least it means we won't crash if
    // we try to read the VBS from this partition!
    if (( LSWAP(pxMBR->aPartitions[i].lOffset)
	 +LSWAP(pxMBR->aPartitions[i].lSize)-1 ) > g_lCardSectors) continue;
    // As far as we can see, the this partition is valid.  Let's use it!
    RELEASE_BUFFER(pBCB);  return LSWAP(pxMBR->aPartitions[i].lOffset);
  }
  
  // Here if no MBR is found - release the disk buffer and return...
FAIL:
  RELEASE_BUFFER(pBCB);  return 0L;
}


//++
//   This routine reads the Volume Boot Sector (aka VBS) from the flash card
// and extracts the information that we'll need in the future to access files
// on this device.  Finding the VBS can be something of an issue - normally
// it's located in sector zero, however if the flash card has been partitioned
// with FDISK then sector zero will contain a Master Boot Record instead.  In
// that case we have to read the partition table in the MBR, select the part-
// ition we're interested in, and use that to find the VBS.
//
//   Note that this routine uses the current directory sector as a temporary
// buffer.  This is clearly harmless, since we can't have a real current
// directory sector before we've read the volume data!
//--
PRIVATE BOOL ReadVBS (LONG lLBN)  STACKSLOCS
{
  PIBCB pBCB;  BOOT_SECTOR XDATA *pxVBS;  BYTE i;

  //   Allocate a temporary buffer for disk I/O.  Once this is done (successfully)
  // we don't dare return without releasing the buffer first!  After that's done,
  // read the boot sector.  
  ALLOCATE_BUFFER(pBCB);
  if (pBCB == NULL) return FALSE;
  //DBGOUT(("ReadVBS: reading volume boot sector from LBN %lu\n", lLBN));
  if (!ReadSector(lLBN, pBCB->pbBuffer)) goto FAIL;
  pxVBS = (BOOT_SECTOR XDATA *) pBCB->pbBuffer;
#ifdef SHOWBOOTSECTOR
  ShowBootSector(pxVBS);
#endif

  //   Sanity check a bunch of stuff in the VBS and BPB to make sure this really
  // is a valid volume boot sector...
  if (WSWAP(pxVBS->wBootSignature) != 0xAA55) {
    DBGOUT(("ReadVBS: Boot signature = 0x%04X (!= 0xAA55)...\n", WSWAP(pxVBS->wBootSignature)));
    goto FAIL;
  } else if (WSWAP(pxVBS->bpb.wBytesPerSector) != IDE_SECTOR_SIZE) {
    DBGOUT(("ReadVBS: Bytes per Sector = %d (!= %d)...\n", WSWAP(pxVBS->bpb.wBytesPerSector), IDE_SECTOR_SIZE));
    goto FAIL;
  } else if (pxVBS->bpb.bMediaDescriptor != 0xF8) {
    DBGOUT(("ReadVBS: Media type = 0x%02bX (!= 0xF8)...\n", pxVBS->bpb.bMediaDescriptor));
    goto FAIL;
  } 

  // Check the file system type to be sure it is FAT16...
  strncpy(g_szSystemType, pxVBS->bpb.szSystemID, SYSTEM_TYPE_LENGTH-1);
  for (i=SYSTEM_TYPE_LENGTH-2;  (i>0) && (g_szSystemType[i]==' ');  --i)  g_szSystemType[i]='\0';      
  if (strncmp(pxVBS->bpb.szSystemID, "FAT16   ", 8) != 0) {
    DBGOUT(("ReadVBS: File system type = %8.8s (!= FAT16)...\n", &(pxVBS->bpb.szSystemID)));
    goto FAIL;
  }

  //  Looks like a valid volume boot sector.  We can use this information to
  // compute the location of the first FAT (the only one we ever use; any
  // seconday FATs are ignored!), the location of the root directory, and the
  // first sector of the user file data area.
  m_lVolFATOffset = lLBN + (LONG) WSWAP(pxVBS->bpb.wReservedSectors);
  m_lVolFATSize = (LONG) pxVBS->bpb.bNumberFATs * (LONG) WSWAP(pxVBS->bpb.wSectorsPerFAT);
  //DBGOUT(("ReadVBS: FAT starts at LBN %ld, length %ld sectors ...\n", m_lVolFATOffset, m_lVolFATSize));
  m_lVolRootOffset = m_lVolFATOffset + m_lVolFATSize;
  m_lVolRootSize = (LONG) WSWAP(pxVBS->bpb.wRootEntries) / DIRECTORY_ENTRIES_PER_SECTOR;
  //DBGOUT(("ReadVBS: Root directory starts at LBN %ld, length %ld sectors ...\n", m_lVolRootOffset, m_lVolRootSize));
  m_lVolDataOffset = m_lVolRootOffset + m_lVolRootSize;
  //DBGOUT(("ReadVBS: User file data starts at LBN %ld ...\n", m_lVolDataOffset));
  m_bVolClusterSize = pxVBS->bpb.bSectorsPerCluster;
  //DBGOUT(("ReadVBS: Sectors per cluster = %bd ...\n", m_bVolClusterSize));

  // Release the buffer and we're done!
  RELEASE_BUFFER(pBCB);  return TRUE;

  //   Here if we fail for any reason - don't you dare return without releasing the
  // disk buffer first!!
FAIL:
  RELEASE_BUFFER(pBCB);  return FALSE;
}


//++
//   This little routine will read the specified directory sector into the
// current directory buffer m_pCurDirBCB.  If the latter is NULL, then it
// will attempt to allocate a buffer first, which will eventually be released
// by FinishDirScan().  If there's an I/O error (unlikely, but possible)
// it will return FALSE...
//--
PRIVATE BOOL ReadDirSector (LONG lSector)
{
  //ASSERT((m_pCurDirBCB != NULL), ("ReadDirectorySector: no current directory BCB!\n"));
  if (m_pCurDirBCB == NULL) {
    ALLOCATE_BUFFER(m_pCurDirBCB);
    if (m_pCurDirBCB == NULL) return FALSE;
    //DBGOUT(("ReadDirSector: allocated buffer 0x%04X\n", (WORD) m_pCurDirBCB->pbBuffer));
  }
  if (!ReadSector(lSector, m_pCurDirBCB->pbBuffer)) {
    DBGOUT(("ReadDirSector: Error reading directory sector %ld\n", lSector));
    return FALSE;
  }
  m_lCurDirSector = lSector;  return TRUE;
  //DBGOUT(("ReadDirSector: read directory sector %ld ...\n", m_lCurDirSector));
}

//++
//   Initialize all the m_?CurDir???? variables so that we can start scanning
// forward from the first entry in the root directory.  The parameter, fForward,
// is TRUE to scan forward from the first file in the directory, and FALSE to
// start scanning backwards from the last file.  The latter is used when the user
// turns the knob counter clockwise, for example.
//--
PRIVATE BOOL StartDirScan (BOOL fForward)
{
  if (fForward) {
    // Read the first sector and start at the beginning...
    if (!ReadDirSector(m_lVolRootOffset)) return FALSE;
    m_bCurDirEntry = 0;
  } else {
    // Read the last sector and start at the end...
    if (!ReadDirSector(m_lVolDataOffset-1)) return FALSE;
    m_bCurDirEntry = DIRECTORY_ENTRIES_PER_SECTOR-1;
  }
  //DBGOUT(("StartDirScan: m_lCurDirSector=%ld, m_bCurDirEntry=%bd, buffer=0x%04X\n",
  //  m_lCurDirSector, m_bCurDirEntry, (WORD) m_pCurDirBCB->pbBuffer));
  return TRUE;
}

//++
//   This routine should be called whenever we've processed all the entries
// in the current directory sector.  It will read the next directory sector
// into m_pCurDirBCB and reset m_bCurDirEntry to zero so that we can continue
// processing the files in the next directory sector.  When we reach the end
// of the root directory, it will return FALSE.
//--
PRIVATE BOOL ReadNextDirSector (void)
{
  m_bCurDirEntry = 0;
  if (m_lCurDirSector >= m_lVolDataOffset-1) return FALSE;
  return ReadDirSector(m_lCurDirSector+1);
}

//++
//   This is essentially the same as ReadNextDirSector() except that it (as if
// you couldn't guess) reads the previous directory segment.  It used for scanning
// backwards, of course - either to find an LFN or when the user turns the knob
// counter clockwise.  It returns FALSE when we reach the _beginnning_ of the
// root directory...
//--
PRIVATE BOOL ReadPrevDirSector (void)
{
  m_bCurDirEntry = DIRECTORY_ENTRIES_PER_SECTOR-1;
  if (m_lCurDirSector <= m_lVolRootOffset) return FALSE;
  return ReadDirSector(m_lCurDirSector-1);
}

//++
//   This routine will release the current directory sector buffer back into
// the free pool.  It'll be re-allocated again the next time ReadDirSector()
// is called.  Needless to say, you need to be careful to make sure m_pCurDirBCB 
// isn't NULL before accessing it!
//--
PRIVATE void ReleaseDirBuffer (void)
{
  if (m_pCurDirBCB != NULL) {
    //DBGOUT(("ReleaseDirBuffer: releasing buffer 0x%04X\n", (WORD) m_pCurDirBCB->pbBuffer));
    RELEASE_BUFFER(m_pCurDirBCB);  m_pCurDirBCB = NULL;
  }
}


//++
//   This routine will return TRUE if the file is an MP3 file, and FALSE
// otherwise.  It also ignores empty files, deleted files, (sub)directory
// files, and volume labels, regardless of their name.  It's used by the
// NextMP3File() and PreviousMP3File() functions to decide what to skip.
//--
PRIVATE BOOL IsMP3File (PXDIRENT pxDirEnt)
{
  if (((BYTE) pxDirEnt->szFileName[0]) == EMPTY_FILE) return FALSE;
  if (((BYTE) pxDirEnt->szFileName[0]) == DELETED_FILE) return FALSE;
  if ((pxDirEnt->bAttributes & (VOLUME_NAME|DIRECTORY_FILE|SYSTEM_FILE|HIDDEN_FILE)) != 0) return FALSE;
  if ( (pxDirEnt->szFileType[0] != 'M')
    || (pxDirEnt->szFileType[1] != 'P')
    || (pxDirEnt->szFileType[2] != '3') ) return FALSE;
  return TRUE;
}

//++
//   This routine will traverse the entire volume root directory and a) count
// the number of MP3 files present, and b) find the volume label file (if any).
// It's admittedly a very special purpose routine (not much a general purpose
// FAT function at all!), but doing it this way allows us to find everything
// we need to know in just one pass thru the directory.  On a volume with lots
// of files, the time savings are important...
//--
PUBLIC void GetVolumeData (void)
{
  PXDIRENT pxDirEnt;
  memset(g_szVolumeLabel, 0, VOLUME_LABEL_LENGTH);  g_wTotalMP3Files = 0;
  if (!StartDirScan(TRUE)) return;

  while (TRUE) {
    // If this is an empty or deleted file, then just skip it....
    pxDirEnt = GetDirEntry(m_bCurDirEntry);
    if (((BYTE) pxDirEnt->szFileName[0]) == EMPTY_FILE) goto NextFile;
    if (((BYTE) pxDirEnt->szFileName[0]) == DELETED_FILE) goto NextFile;

    //   If it's a volume label, then save the name in g_szVolumeLabel and go
    // on to the next file.  Notice that we have to be careful here, because a
    // file with both the VOLUME LABEL and the DIRECTORY attributes set is an
    // LFN entry, and those are common!
    if (pxDirEnt->bAttributes == VOLUME_NAME) {
      //DBGOUT(("GetVolumeData: volume label = \"%8.8s%3.3s\" attr=0x%02bX\n",
      //   pxDirEnt->szFileName, pxDirEnt->szFileType, pxDirEnt->bAttributes));
      strncpy(g_szVolumeLabel, pxDirEnt->szFileName, VOLUME_LABEL_LENGTH-1);
      goto NextFile;
    }

    // If it's an MP3 file, count it.  Otherwise, ignore it!
    if (IsMP3File(pxDirEnt)) ++g_wTotalMP3Files;

    // Find the next file in the directory....
NextFile:
    if (++m_bCurDirEntry >= DIRECTORY_ENTRIES_PER_SECTOR) {
      if (!ReadNextDirSector()) break;
    }
  }

  ReleaseDirBuffer();
}


//++
//   This routine will attempt to get the LFN (that's Long File Name, if you've
// been out of touch for a while :-) that corresponds to the directory entry
// specified.  This is a slight kludge because it works only for the current
// directory entry - the one just returned by Prev/NextMP3File().  You can't
// save away a directory entry and then call this routine some time later.
// Sorry, but that's not a problem for the player.
//
//   LFNs are stored in bogus directory entries that appear just BEFORE the
// real directory entry.  Unfortunatly there's no guarantee that the entire
// LFN will appear in one sector, so it may be necessary for us to back up
// and re-read the previous directory sector to complete the LFN. 
//
//   Finally, LFNs are stored in Unicode with 16 bit characters.  We convert
// these to ASCII by the simple expedient of dropping the upper byte - that
// mostly works for all Western character sets...
//--
PRIVATE BOOL GetFileLFN (void) STACKSLOCS
{
  BYTE bSequence, bEntry, i, cbString;  LONG lOriginalDirectorySector;
  BOOL fLFNFound;  PXSTRING pxString;  LFN_DIRECTORY_ENTRY XDATA *pxLFN;
  
  ASSERT( (m_lCurDirSector != 0), ("Call NextMP3File() first!") );
  memset(g_szLongFileName, 0, MAX_LFN_LENGTH);  pxString = g_szLongFileName;
  cbString = MAX_LFN_LENGTH;  bSequence = 1; fLFNFound = FALSE;
  lOriginalDirectorySector = m_lCurDirSector;  bEntry = m_bCurDirEntry;
  //DBGOUT(("GetFileLFN: searching for LFN for file \"%8.8s.%3.3s\"\n",
  //  &(GetCurDirEntry()->szFileName), &(GetCurDirEntry()->szFileType) ));

  //   Start with the previous directory entry and keep reading backwards
  // (possibly re-reading the previous sector, if required) until we find
  // something which is NOT a LFN entry.
  while (TRUE) {
    if (bEntry == 0) {
      // Re-read the previous directory sector...
      if (m_lCurDirSector == m_lVolRootOffset) break;
      --m_lCurDirSector;  bEntry = DIRECTORY_ENTRIES_PER_SECTOR - 1;
      if (!ReadDirSector(m_lCurDirSector)) break;
    } else
      --bEntry;
    //   See if this entry is another LFN segment.  Note that LFN entries
    // actually have a checksum, which we could (and probably should!) actually
    // check to make sure this entry is valid.  We don't bother, however, because
    // the other checks are usually enough...
    pxLFN = (LFN_DIRECTORY_ENTRY XDATA *) GetDirEntry(bEntry);
    //DBGOUT(("GetFileLFN: checking, Attributes=0x%02bX, Sequence=%bd, MBZ1,2=%bd %u\n",
    //  pxLFN->bAttributes, pxLFN->bSequence, pxLFN->bMBZ1, pxLFN->wMBZ2 ));
    if (pxLFN->bAttributes != LFN_ATTRIBUTES) break;
    if ((pxLFN->bSequence & 0x3F) != bSequence) break;
    if ((pxLFN->bMBZ1 != 0) || (pxLFN->wMBZ2 != 0)) break;
    //   Ok, this looks like a real LFN entry.  Extract the characters from it
    // and store them in the caller's buffer.  Remember that Unicode character
    // 0xFFFF is used to signal the end of the string.  BTW, using the sequence
    // LOBYTE(WSWAP(x)) is kind of wasteful (and it's a little too much for C51
    // to optimize away) but it is portable.  You can't just replace it with
    // HIBYTE(x) because SDCC is little endian and there WSWAP() is a no-op.
    //
    //   Lastly, you might think it would be more effective to write a sub-
    // routine for this, but it would require pointers to awChars, pxString
    // (the variable itself, so it can be updated), and cbMaxString (again,
    // so that it can be updated) at very least.  It actually generates less
    // code on the 8051 to just write it out the long way.
    for (i = 0;  i < 5;  ++i) {
      if (pxLFN->awChars1[i] == 0xFFFF) goto EndOfLFN;
      if (--cbString == 0) goto EndOfLFN;
      *pxString++ = LOBYTE(WSWAP(pxLFN->awChars1[i]));
    }
    for (i = 0;  i < 6;  ++i) {
      if (pxLFN->awChars2[i] == 0xFFFF) goto EndOfLFN;
      if (--cbString == 0) goto EndOfLFN;
      *pxString++ = LOBYTE(WSWAP(pxLFN->awChars2[i]));
    }
    for (i = 0;  i < 2;  ++i) {
      if (pxLFN->awChars3[i] == 0xFFFF) goto EndOfLFN;
      if (--cbString == 0) goto EndOfLFN;
      *pxString++ = LOBYTE(WSWAP(pxLFN->awChars3[i]));
    }
    // Remember that we found an LFN and bump the sequence for the next one...
    fLFNFound = TRUE;  ++bSequence;
    //   If bit 0x40 if the sequence number is set, then this is the last
    // entry in this LFN and we can stop. 
    if ((pxLFN->bSequence & 0x40) != 0) goto EndOfLFN;
  }
EndOfLFN:

  //   If we've changed the directory sector, then be sure and restore the
  // original before we leave...
  if (m_lCurDirSector != lOriginalDirectorySector) {
    m_lCurDirSector = lOriginalDirectorySector;  ReadDirSector(m_lCurDirSector);
  }
#ifdef DEBUG
  if (fLFNFound) DBGOUT(("GetFileLFN: \"%s\"\n", g_szLongFileName));
#endif
  return fLFNFound;
}


//++
//   This routine will return the first MP3 file found in the root directory.
// Normally it's called once, when the player starts up, and then is followed
// by many calls to NextMP3File() or PrevMP3File().  It goes something like
// this:
//
//	PXDIRENT pxFile = FirstMP3File();
//	...
//	while (pxFile != NULL) {
//	  ...
//	  pxFile = NextMP3File();
//	  ... or ...
//	  pxFile = PrevMP3File();
//	}
//
//   This loop will normally execute endlessly playing the all the MP3 files in
// the root directory over and over again.  pxFile will be NULL only in the case
// of severe errors (i.e. a hardware failure reading the directory) - in all other
// cases, NextMP3File() or PrevMP3File() will loop back to the beginning or end of
// the root directory as required.  Perhaps not the most useful general purpose
// directory traversal code, but just what we want for an MP3 player!
//--
PUBLIC PXDIRENT FirstMP3File (void)
{
  ASSERT((m_lVolRootOffset != 0), ("Call InitializeFAT() first!"));
  if (!StartDirScan(TRUE)) return NULL;  
  return NextMP3File();
}

//++
//   This routine searches the current directory for the next MP3 file.
// The m_lCurDirSector and m_bCurDirEntry globals should already be set up
// and optionally, the current directory segment can be in memory in the
// buffer addressed by m_pCurDirBCB.  However if the latter is NULL, this
// routine will re-read the current directory sector before proceeding;
// this allows us to return the directory sector buffer to the free list
// while we're actually playing an MP3 file, and that gives us an extra
// buffer for MP3 data.
//
//   This routine automatically skips empty files, deleted files, any
// volume name file, any subdirectory file, and any file marked as system
// or hidden.  All other files, including read only files, will be returned.
// It's up to the caller to sort out the ones the application wants.
//
//   N.B.  When the end of the directory is reached, this routine simply
// loops back to the start and returns the same set of files all over again.
// It returns NULL only in the case of gross failure (e.g. CompactFlash read
// errors).  It's not that useful for general directory traversal, but it's
// just the thing for an MP3 player!!
//
//   Notice also that this will unfortunately loop forever (!!) if there are
// no MP3 files in the root directory of this volume, but we know that's not
// the case because the code has already checked this by calling GetVolumeData().
//
//   And one final word - note that return return value is supposed to be a
// pointer to the current entry in the directory buffer.  "Supposed to be", but
// it isn't really because we release the directory buffer while we're playing
// MP3s.  Instead, this routine makes a copy of just the current directory
// entry into a temporary buffer in XDATA and returns the address of that.
// The caller can't tell the difference either way...
//--
PUBLIC PXDIRENT NextMP3File (void)
{
  PXDIRENT pxDirEnt;
  //  If the directory sector buffer has been released, then re-read the
  // current directory sector before proceeding!
  if (m_pCurDirBCB == NULL) ReadDirSector(m_lCurDirSector);
  while (TRUE) {
    //   Advance m_bCurDirEntry to the next entry in this sector.  If we've
    // reached the end, then read the next sector.  If that fails (because
    // we've reached the end of the root directory) then start over again.
    if (++m_bCurDirEntry >= DIRECTORY_ENTRIES_PER_SECTOR)
      if (!ReadNextDirSector())
	if (!StartDirScan(TRUE)) return NULL;

    // We've got a file - see if it's an MP3 file ...
    pxDirEnt = GetDirEntry(m_bCurDirEntry);
    if (IsMP3File(pxDirEnt)) {
      //   Read the LFN for this file (if it has one), which we have to do while
      // the current directory sector is still in memory, and then cache the
      // file's information.  Release the directory segment, and return a pointer
      // to the cached file information...
      GetFileLFN();  memcpy(&m_CurDirEntCache, pxDirEnt, sizeof(DIRECTORY_ENTRY));
      ReleaseDirBuffer();  return &m_CurDirEntCache;
    }
  }
}

//++
//   This routine is essentially the same as NextMP3File(), except that it goes
// backwards rather than forwards.  It's used by the navigation knob to find the
// previous MP3 file.  When we reach the beginning of the directory, it loops
// back to the end and starts returning the same list of files all over again.
//--
PUBLIC PXDIRENT PrevMP3File (void)
{
  PXDIRENT pxDirEnt;
  if (m_pCurDirBCB == NULL) ReadDirSector(m_lCurDirSector);
  while (TRUE) {
    if (m_bCurDirEntry-- == 0)
      if (!ReadPrevDirSector())
	if (!StartDirScan(FALSE)) return NULL;
    pxDirEnt = GetDirEntry(m_bCurDirEntry);
    if (IsMP3File(pxDirEnt)) {
      GetFileLFN();  memcpy(&m_CurDirEntCache, pxDirEnt, sizeof(DIRECTORY_ENTRY));
      ReleaseDirBuffer();  return &m_CurDirEntCache;
    }
  }
}


//++
// Return the starting sector number for a specific cluster...
//--
PRIVATE LONG SectorFromCluster (CLUSTER wCluster)
{
  LONG lSector;
  lSector = ((LONG) (wCluster-CLUSTER_OFFSET)) * ((LONG) (m_bVolClusterSize)) + m_lVolDataOffset;
  //DBGOUT(("SectorFromCluster: Cluster %u mapped to sector %ld ...\n", wCluster, lSector));
  return lSector;
}

//++
//   The first cluster in a file is easily determined from its directory entry.
// If a file consists of more than one cluster, then the FAT entry for the first
// cluster contains the number of the next cluster; the FAT entry for the second
// cluster points to the third, and so on.  Given the current cluster in a file,
// this function will determine the next cluster, reading the FAT as necessary.
// Note that this depends solely on the current cluster; nothing else about the
// file matters (at least not here!).
//
//   Notice that the current FAT sector is kept in the buffer pointed to by
// m_pCurFATBCB.  Unlike the directory buffer, which is released while we're
// actually playing an MP3 file, the current FAT sector is ALWAYS kept in memory
// even while we're playing.  In principle we could release the FAT buffer too,
// but remember that we need to keep calling NextCluster() to read the MP3 file
// even while we're playing, and that would mean that we'd need to keep re-reading
// the current FAT sector while we play.  That's just too much overhead!
//--
PRIVATE WORD NextCluster (CLUSTER wCluster)
{
  LONG lFATSector;  BYTE bFATOffset; 
  //   The first step is to figure out which FAT location, both the FAT sector and
  // the offset within that sector, holds the entry for the current cluster.
  // IMPORTANT!  Note that right now, IDE_SECTOR_SIZE == 512, sizeof(CLUSTER) == 2,
  // and CLUSTERS_PER_SECTOR == 256.  So long as this is true (and it probably will
  // be until somebody implements FAT12 or FAT32), we can save ourselves a lot of
  // time here, especially on a microprocessor where long divisions are expensive!
  //lFATSector = (WCluster) / CLUSTERS_PER_SECTOR;
  //lFATSector = (wCluster) % CLUSTERS_PER_SECTOR;
  lFATSector = m_lVolFATOffset + HIBYTE(wCluster);
  bFATOffset = LOBYTE(wCluster);

  // If the required FAT sector isn't already in memory, then read it...
  ASSERT((lFATSector != 0),("FAT sector cannot be zero!"));
  if (lFATSector != m_lCurFATSector) {
    if (!ReadSector(lFATSector, (PXBYTE) &m_awCurFAT)) {
      DBGOUT(("NextCluster: Error reading FAT sector %ld\n", lFATSector));
      return EOF_CLUSTER;
    }
    //DBGOUT(("NextCluster: Read FAT sector %ld ...\n", lFATSector));
    m_lCurFATSector = lFATSector;
  }

  // The rest is easy!
  return WSWAP(m_awCurFAT[bFATOffset]);
}


//++
//   This routine will set up the current file data (m_?CurFile???) to the beginning
// of the file entry specified.  The code can then read this file, one sector at
// a time, by calling ReadFile().  Note that this code isn't re-entrant, and there
// is only one set of current file data, so only one file can be opened for reading
// at a time.  In an MP3 player, that isn't a problem...
//--
PUBLIC void OpenFile (PXDIRENT pxDirEnt)
{
  LONG lTemp;
  m_wCurFileCluster = WSWAP(pxDirEnt->wFirstCluster);
  m_lCurFileSector = SectorFromCluster(m_wCurFileCluster);
  m_nRelFileSector = -1;
  lTemp = pxDirEnt->lFileSize;  m_lFileBytesLeft = LSWAP(lTemp);
  //DBGOUT(("OpenFile: reading file \"%8.8s.%3.3s\",", &(pxDirEnt->szFileName), &(pxDirEnt->szFileType)));
  //DBGOUT((" starting LBN=%ld, file size=%ld\n", m_lCurFileSector, m_lFileBytesLeft));
}

//++
//   This routine will find the next sector in the currently open file.  If
// all the sectors in the current cluster have been read, then the FAT will
// be consulted to find the next sector in the file.  If there are no more
// clusters (i.e. the FAT entry is EOF_CLUSTER), then FALSE is returned.
// This last condition can't be used as an indication of end-of-file, however,
// because files rarely if ever exactly fill a cluster.  The application must
// keep count of the number of bytes read and use the file size from the
// directory entry to determine EOF instead.
//--
PRIVATE BOOL NextFileSector (void)
{
  //   Note that OpenFile() sets nRelSector to -1, so the first time this
  // routine is called we'll end up reading the first sector in the file...
  if (m_wCurFileCluster == EOF_CLUSTER) return FALSE;
  if (++m_nRelFileSector >= m_bVolClusterSize) {
    m_nRelFileSector = 0;
    m_wCurFileCluster = NextCluster(m_wCurFileCluster);
    if (m_wCurFileCluster == EOF_CLUSTER) return FALSE;
    m_lCurFileSector = SectorFromCluster(m_wCurFileCluster);
  }
  return TRUE;
}

//++
//   This routine reads the currently opened file (remember that there can be
// only one file open at a time!) sector by sector.  The pxBuffer parameter
// should point to a 512 byte buffer, and it will return the number of bytes
// actually read.  Normally this will be 512, however it'll be less when we
// reach the last block in the file.
//--
PUBLIC WORD ReadFile (PXBYTE pxBuffer)
{
  LONG lSector;
  // If there are no bytes left in this file, then we always fail...
  if (m_lFileBytesLeft == 0) return 0;

  // We should never find EOF before the byte count reaches zero!
  VERIFY( (NextFileSector()), ("ReadFile: EOF_CLUSTER found before byte count = 0\n") );

  //   Now we can read the sector.  If this fails for some reason (it shouldn't!)
  // the return zero and fake EOF...
  lSector = m_lCurFileSector + (LONG) (m_nRelFileSector);
  if (!ReadSector(lSector, pxBuffer)) {
    DBGOUT(("ReadFile: error reading sector %ld from current file!\n", lSector));
    return 0;
  }

  //   We've read another sector, so decrement the number of bytes left in the
  // file accordingly.  If the number of bytes left is currently less than 512,
  // then we've only read a partial sector (the last sector of the file wasn't
  // full) and we have to zero the bytes left and return the appropriate count.
  if (m_lFileBytesLeft >= IDE_SECTOR_SIZE) {
    m_lFileBytesLeft -= IDE_SECTOR_SIZE;  return IDE_SECTOR_SIZE;
  } else {
    LONG lTemp = m_lFileBytesLeft;  m_lFileBytesLeft = 0;
    return lTemp;
  }
}

//++
//   This routine will read the last cbBuffer bytes from the end of the currently
// opened file.  Ordinarily you'd do an lseek() and then a read(), but, of course,
// we have no lseek() here.  This is a rather esoteric function, but the purpose
// is obvious - it's used to read ID3 tags!
//
//   The bad thing about a FAT file system is that there's no way to find the last
// cluster in a file except by following the entire cluster chain from the start. 
// Unfortunately, there's just no way to make this efficient.  With FAT16 a typical
// 512Mb CompactFlash card will have 8K byte sectors (16K byte for a 1Gb card) and
// at that rate, a typical 2 or 3Mb MP3 file will contain 300 to 400 clusters.  With
// hundreds of clusters involved, finding the end of the file can take a long time.
//
//   Notice that we need a temporary disk buffer for reading sectors from the MP3
// file.  We pull one off the free buffer list, but we're careful to always return
// it before leaving!
//--
PUBLIC BOOL ReadFileEnd (PXBYTE pxBuffer, WORD cbBuffer) STACKSLOCS
{
  WORD wCount;  LONG lClusterSize, lSizeLimit;  PIBCB pBCB;
  ALLOCATE_BUFFER(pBCB);
  if (pBCB == NULL) return FALSE;

  //   First, we want to get as close as we can to the end of the file with
  // out actually passing the ID3 tag.  Remember that there's a possibility
  // that the ID3 tag might be split across the last two sectors in the file,
  // and if that happens there's an even smaller possibility that the last
  // two sectors might be in different clusters!  But we know that as long as
  // there are at least the cluster size plus the tag size bytes left in the
  // file, there's no way we could have passed it.
  //
  //   The lSizeLimit variable is just an optimization because the compiler
  // isn't smart enough to factor out the calculation from the while loop.
  // Remember that on an 8051, LONGs are quadruple precision arithmetic!
  lClusterSize = (LONG) m_bVolClusterSize * (LONG) IDE_SECTOR_SIZE;
  lSizeLimit = lClusterSize + cbBuffer;
  while (m_lFileBytesLeft >= lSizeLimit) {
    m_wCurFileCluster = NextCluster(m_wCurFileCluster);
    ASSERT( (m_wCurFileCluster != EOF_CLUSTER), ("ReadFileEnd: Premature EOF cluster\n") );
    m_lFileBytesLeft -= (LONG) lClusterSize;
  }
  //  Be sure that the rest of the m_CurFile block, especially the current
  // sector information, is valid before we proceed.  Otherwise ReadFile()
  // will crash and burn horribly...
  m_nRelFileSector = -1;
  m_lCurFileSector = SectorFromCluster(m_wCurFileCluster);

  //   Now find the last (or the second to the last, if the tag is split) sector
  // in the file.  This loop is even slower than the last one, but unlike the
  // last one, which can execute hundreds of times, this one never executes more
  // than fifteen times (assuming 8K clusters).
  while (m_lFileBytesLeft >= (IDE_SECTOR_SIZE+cbBuffer)) {
    VERIFY ( (NextFileSector()), ("ReadFileEnd: Premature EOF sector\n") );
    m_lFileBytesLeft -= IDE_SECTOR_SIZE;
  }

  // Ok, we're there.  Read at least the first part of the tag into the buffer.
  wCount = ReadFile(pBCB->pbBuffer);
  ASSERT( (wCount > 0), ("ReadFileEnd: Failed to read last chunk in file\n") );
  if (wCount == 0) {RELEASE_BUFFER(pBCB);  return FALSE;}

  //   If that last ReadFile() reached the end of the file, then we know
  // that we have the ID3 tag in the buffer now.  All we need to do is find
  // it and copy it to the ID3 tag buffer.  However, if there are still bytes
  // left in the file then we know we have only part of the tag and we must
  // do another ReadFile() to get the rest.
  if (m_lFileBytesLeft == 0) {
    // The easy case - we got the whole tag in one sector...
    ASSERT( (wCount >= cbBuffer), ("ReadFileEnd: wCount < cbBuffer\n") );
    memcpy(pxBuffer, pBCB->pbBuffer+(wCount-cbBuffer), cbBuffer);
  } else {
    //   This tag is split across the last two sectors in the file.  Figure out
    // how many bytes (of the tag) we got in this sector and extract those first.
    WORD wFragmentSize;
    //DBGOUT(("ReadFileEnd: fragmented, lBytesLeft=%ld\n", m_lFileBytesLeft));
    ASSERT( (m_lFileBytesLeft < cbBuffer), ("ReadFileEnd: lBytesLeft >= cbBuffer\n") );
    wFragmentSize = cbBuffer - m_lFileBytesLeft;
    memcpy(pxBuffer, pBCB->pbBuffer+IDE_SECTOR_SIZE-wFragmentSize, wFragmentSize);
    //   And then do another read to get the rest.  Remember that since the
    // ID3 tag is the last thing in the file, ALL the bytes returned by this
    // second ReadFile() must be part of the tag. 
    wCount = ReadFile(pBCB->pbBuffer);
    ASSERT( (wCount > 0), ("ReadFileEnd: Failed to read part 2\n") );
    ASSERT( (m_lFileBytesLeft == 0), ("ReadFileEnd: Second read didn't finish file\n") );
    memcpy(pxBuffer+wFragmentSize, pBCB->pbBuffer, wCount);
  }

  // We can give up the temporary buffer now ....
  RELEASE_BUFFER(pBCB);
  return TRUE;
}

//++
//   This routine exists only for completeness in abstracting the file system
// layer.  In this particular implementation, it doesn't do much.
//--
PUBLIC void CloseFile (void)
{
  m_wCurFileCluster = 0;  m_nRelFileSector = 0;
  m_lCurFileSector = m_lFileBytesLeft = 0;
  //DBGOUT(("CloseFile: ...\n"));
}


//++
//   This routine reads the current compact flash card to see if it contains
// a valid FAT16 file system and, if it does, it sets up the volume data
// globals with information we'll need to access this volume in the future.
// If the current CF card DOESN'T contain a FAT16 file system, then this
// function returns FALSE...
//--
PUBLIC BOOL InitializeFAT (void)
{
  //   Initialize all the local data structures.  Notice that this code is a
  // little careless because it simply initializes both m_pCurDirBCB and
  // m_pCurFATBCB to NULL.  It's important (critical!) that both these be
  // initialized, but if InitializeFAT() were ever to be called more than
  // once (say, if we actually supported hot swapping CF cards) then this
  // would cause a big time memory leak since we'd just be throwing away
  // any buffer that had been previously allocated.  At the moment, though,
  // InitializeFAT() is only ever called once, so this isn't a problem.
  memset(g_szSystemType, 0, SYSTEM_TYPE_LENGTH);
  memset(g_szVolumeLabel, 0, VOLUME_LABEL_LENGTH);
  g_wTotalMP3Files = 0;
  // Reset all the m_?VolData* variables ...
  m_lVolFATOffset = m_lVolRootOffset =  m_lVolDataOffset = 0;
  m_lVolFATSize = m_lVolRootSize = 0;  m_bVolClusterSize = 0;
  // Reset all the m_?CurDir* variables ...
  m_lCurDirSector = 0;  m_pCurDirBCB = NULL;
  m_bCurDirEntry = DIRECTORY_ENTRIES_PER_SECTOR;
  // Reset all the m_?CurFAT* variables ...
  m_lCurFATSector = 0;  memset(m_awCurFAT, 0, sizeof(m_awCurFAT));
  // Reset all the current file variables...
  m_wCurFileCluster = 0;  m_nRelFileSector = 0;
  m_lCurFileSector = m_lFileBytesLeft = 0;

  //  The next step is to find the volume boot sector (aka VBS).  Experimentally
  // is seems like Windows formats any removable media device, regardless of the
  // size, as if it were a big floppy disk.  This means that the VBS goes in
  // sector zero of the device.  However, there's nothing that prevents someone
  // from running FDISK on a compact flash drive, and will write a Master Boot
  // Record (aka MBR) to sector zero instead.  In this case the MBR contains a
  // partition table, the partition table points to the first sector of each
  // logical volume, and THAT sector then contains the VBS (there's a separate
  // VBS for every partition).
  if (!ReadVBS(ReadMBR())) return FALSE;
  DBGOUT(("Found file system type \"%s\" ...\n", g_szSystemType));

  return TRUE;
}



