	.TITLE	(STA013 Data Request Interrupt Service)
	.sbttl	Copyright (C) 2005 by Spare Time Gizmos.  All rights reserved.


;++
; mp3drq.a51
;
; Copyright (C) 2005 by Spare Time Gizmos.  All rights reserved.
;
; This file is part of the Spare Time Gizmos' MP3 Player firmware.
;
; This firmware is free software; you can redistribute it and/or modify it under
; the terms of the GNU General Public License as published by the Free Software
; Foundation; either version 2 of the License, or (at your option) any later
; version.
;
; This program is distributed in the hope that it will be useful, but WITHOUT
; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
; more details.
;
; You should have received a copy of the GNU General Public License along with
; this program; if not, write to the Free Software Foundation, Inc., 59 Temple
; Place, Suite 330, Boston, MA  02111-1307  USA
; 
; DESCRIPTION:
;   This module contains the STA013 data request (MP3 data, that is) interrupt
; service.  The STA013 wants its MP3 data delivered bit serially with a synch-
; ronous clock.  In DEBUG mode we do this the hard way, by bit banging two
; port pins.  Needless to say, this is a little slow and limits the bit rate
; of the MP3s we can play to around 32kbps or so.
; 
;   In non-DEBUG mode we can use the 8051's built in UART in mode 0 to shift
; out the bits; this is vastly faster and is able to play 128kbps MP3s without
; missing a beat (literally!), however there is a small downside.  The 8051's
; UART sends data LSB first, which is the opposite of the way the STA013 wants
; to see it.  There's no choice but to reverse the order of the bits in each
; byte - to prevent this expensive operation from eating up all the time we
; save, we use a 256 byte lookup table to handle the reversal.
; 
;   The MP3 file data is buffered in a very simple circular buffer stored in
; xdata RAM.  In practice this buffer is always an exact multiple of 512 bytes
; so that IDE sectors can be read directly without any nasty wrap around 
; problems, but this code actually doesn't care what size the buffer is.
; 
; REFERENCES:
; 	http://www.pjrc.com/tech/mp3/sta013.html
;
; REVISION HISTORY:
; dd-mmm-yy     who     description
; 29-May-05	RLA	New File
; 12-Jul-05	DJA	ported to ASX8051/SDCC
; 13-Jul-05	RLA	Unfortunately you can't test for the play list empty
;			  simply by saying "if (g_pPlayListHead == NULL)"
;			  because MP3DRQ always holds the last BCB pointer in
;			  R0.  Invent IsPlayListEmpty() to fix this problem.
; 20-Jul-05	DJA	Port attempt II
; 20-Oct-05	RLA	Correct for SDCC's little endian byte ordering
;		RLA	Missing a # (immedate mode) when setting SCON!
;--

	.module	MP3DRQ
	.optsdcc -mmcs51 --model-small
	.radix	 d
	.include "player.inc"
	.include "fastsdi.inc"

	.globl	_InitializeMP3DRQ, _FlushPlayList, _IsPlayListEmpty
	.globl	_g_pPlayListHead, _g_pPlayListEnd, _MP3DRQ

;   These two global variables keep track of the current queue of MP3 data buffers
; to be played.  The background code pulls buffers off the free list, reads data
; blocks from the CF card, and puts full buffers on this play list.  This code pulls
; a full buffer off the list, sends it to the STA013, and then puts the buffer
; back on the free list.
	.area	DSEG (DATA)
_g_pPlayListHead:	; PUBLIC BUFFER_CONTROL_BLOCK idata *g_pPlayListHead;
	.ds	1
_g_pPlayListEnd:	; PUBLIC BUFFER_CONTROL_BLOCK idata *g_pPlayListEnd;
	.ds	1	


;   Not only do we need the register banks defined, but in this case we
; actually need _two_ banks, zero and one.  To avoid confusion, we actually
; define different (abeit non-standard) names for each bank...  See my
; insightful comments in the ide.asm module for more details...
	.area	REG_BANK_0 (REL,OVR,DATA)
	.ds	8
AR00=0x00
AR01=0x01
AR02=0x02
AR03=0x03
AR04=0x04
AR05=0x05
AR06=0x06
AR07=0x07
	.area	REG_BANK_1 (REL,OVR,DATA)
	.ds	8
AR10=0x08
AR11=0x09
AR12=0x0A
AR13=0x0B
AR14=0x0C
AR15=0x0D
AR16=0x0E
AR17=0x0F

	.area	CSEG (CODE)

;   When we're using the 8051 UART to send data to the STA013, we need to reverse
; the order of the bits in each byte first.  This 256 byte table does the job
; without wasting lots of valuable CPU time (e.g. m_abReverseBits[i] = <the reverse
; of i>).
	.if	FAST_SDI
m_abReverseBits:
;		 00   01   02   03   04   05   06   07   08   09   0a   0b   0c   0d   0e   0f
	.db	0x00,0x80,0x40,0xc0,0x20,0xa0,0x60,0xe0,0x10,0x90,0x50,0xd0,0x30,0xb0,0x70,0xf0 ; 00
	.db	0x08,0x88,0x48,0xc8,0x28,0xa8,0x68,0xe8,0x18,0x98,0x58,0xd8,0x38,0xb8,0x78,0xf8 ; 10
	.db	0x04,0x84,0x44,0xc4,0x24,0xa4,0x64,0xe4,0x14,0x94,0x54,0xd4,0x34,0xb4,0x74,0xf4 ; 20
	.db	0x0c,0x8c,0x4c,0xcc,0x2c,0xac,0x6c,0xec,0x1c,0x9c,0x5c,0xdc,0x3c,0xbc,0x7c,0xfc ; 30
	.db	0x02,0x82,0x42,0xc2,0x22,0xa2,0x62,0xe2,0x12,0x92,0x52,0xd2,0x32,0xb2,0x72,0xf2 ; 40
	.db	0x0a,0x8a,0x4a,0xca,0x2a,0xaa,0x6a,0xea,0x1a,0x9a,0x5a,0xda,0x3a,0xba,0x7a,0xfa ; 50
	.db	0x06,0x86,0x46,0xc6,0x26,0xa6,0x66,0xe6,0x16,0x96,0x56,0xd6,0x36,0xb6,0x76,0xf6 ; 60
	.db	0x0e,0x8e,0x4e,0xce,0x2e,0xae,0x6e,0xee,0x1e,0x9e,0x5e,0xde,0x3e,0xbe,0x7e,0xfe ; 70
	.db	0x01,0x81,0x41,0xc1,0x21,0xa1,0x61,0xe1,0x11,0x91,0x51,0xd1,0x31,0xb1,0x71,0xf1 ; 80
	.db	0x09,0x89,0x49,0xc9,0x29,0xa9,0x69,0xe9,0x19,0x99,0x59,0xd9,0x39,0xb9,0x79,0xf9 ; 90
	.db	0x05,0x85,0x45,0xc5,0x25,0xa5,0x65,0xe5,0x15,0x95,0x55,0xd5,0x35,0xb5,0x75,0xf5 ; a0
	.db	0x0d,0x8d,0x4d,0xcd,0x2d,0xad,0x6d,0xed,0x1d,0x9d,0x5d,0xdd,0x3d,0xbd,0x7d,0xfd ; b0
	.db	0x03,0x83,0x43,0xc3,0x23,0xa3,0x63,0xe3,0x13,0x93,0x53,0xd3,0x33,0xb3,0x73,0xf3 ; c0
	.db	0x0b,0x8b,0x4b,0xcb,0x2b,0xab,0x6b,0xeb,0x1b,0x9b,0x5b,0xdb,0x3b,0xbb,0x7b,0xfb ; d0
	.db	0x07,0x87,0x47,0xc7,0x27,0xa7,0x67,0xe7,0x17,0x97,0x57,0xd7,0x37,0xb7,0x77,0xf7 ; e0
	.db	0x0f,0x8f,0x4f,0xcf,0x2f,0xaf,0x6f,0xef,0x1f,0x9f,0x5f,0xdf,0x3f,0xbf,0x7f,0xff ; f0
	.endif


;++
; PRIVATE void MP3DRQ (void) interrupt 0 using 1
;
;   As you've probably figured out by now, this routine is the interrupt service
; for STA013 MP3 data requests.  In twenty words or less, what we want to do is
; to pull a buffer of MP3 data from the m_pPlayList and send it, synchronous bit
; serially, to the STA013.  When we've finished sending the buffer we return it
; to the free buffer pool, m_pFreeBufferList, and then go back to m_pPlayList to
; look for another buffer to play.
;
;   There are only two things that make our life difficult - the first is that
; the STA013's internal buffer can only hold 200 bytes or so (the datasheet is
; a little vague on that topic!) where as our buffers hold 512.  That means it'll
; take several MP3 DRQ interrupts to transfer a single buffer.
;
;   Second, if the background code is too slow it's possible that we'll run out
; of buffers on the m_pPlayList.  We can't simply return if that happens because
; the STA013 will continue to assert DRQ which will continue to hold the 8051's
; INT0 low.  Since INT0 is configured as level sensitive, we'll never get out of
; this interrupt routine!  The only solution is to mask INT0 interrupts with the
; EX0 bit when this happens.  The background code will turn EX0 back on when it
; reads the next buffer.
;
;   This routine has its own register bank allocated to it, so when we enter
; routine we can depend on R0..R7 having the same values we left there the last
; time we were here.  In general, the register usage is:
;
;	R0 - points to current buffer control block (BCB)
;	R1 - temporary (copy of R0, and loop counter)
;	R2
;	R3
;	R4 - count of bytes remaining in current buffer (high)
;	R5 -  "  "  "  "     "    "   "   "   "  "   "  (low)
;	R6 - address (xdata) of next byte in current buffer (DPH)
;	R7 -  "   "     "    "    "   "   "    "  "   "  "  (DPL)
;--

;   Unlike A51, with sdcc/asx8051 there's no need (and indeed, no way!) to set
; the interrupt vector here to point to MP3DRQ.  The SDCC compiler will take
; care of it for us in the "main" module (player.c in this case) SO LONG AS
; we include a function prototype for the MP3DRQ routine.  The compiler looks
; for the "interrupt" attribute on the prototype declaration and automatically
; generates the appropriate vector.

;   A further unpleasant business is that asx8051 has nothing equivalent to
; the "USING" directive.  The only way to handle different register banks is
; to actually equate the AR0..7 symbols to different addresses.  Since this
; module actually uses both bank zero (for code that's called directly from
; C) and bank one (for this ISR) we actually define a different set of symbols
; (e.g. AR01 for register 1, bank 0 - AR10 for register 0, bank 1) to avoid
; confusion...  Of course, it's still up to us, the programmers, to ensure
; that the correct bank is selected by the PSW at the correct time!

_MP3DRQ:
	PUSH 	ACC			; save the accumulator
	PUSH 	DPH			;  ... data pointer
	PUSH 	DPL			;  ...
	PUSH 	PSW			; register bank and flags
	SETB 	LED_BIT			; (turn the LED on for timing purposes)
	MOV  	PSW, #0x08		; and select register bank #1

;   If there's no current buffer, try to remove the next one from the
; play list.  If the play list is empty, then disable the DRQ interrupts
; and wait for more to arrive.  As long as we leave with R0=0, we'll try
; to obtain a play buffer next time we come here.
ISR10:	CJNE	R0, #0, ISR20		; jump if the current BCB is not null
	MOV  	R0, _g_pPlayListHead	; remove the next BCB from the play list
	CJNE	R0, #0, ISR11		; jump if we really got a buffer
	CLR  	EX0			; no more buffers - disable future interrupts
	SJMP 	ISR99			; and go dismiss this one
;   R0 contains the address of the next buffer on the play list.  Update
; m_pPlayListHead to remove that buffer from the list...
ISR11:	MOV  	_g_pPlayListHead, @R0	;  ... m_pPlayListHead = m_pPlayListHead->pNext
	MOV	R1, AR10		; copy R0 so we don't have to restore it later
;   Load R6/R7 with the address (in xdata) of the actual data buffer that goes
; with this BCB, and load R4/R5 with the count of bytes (up to 512) in this buffer.
	INC	R1			; pbBuffer is the second field in the BCB
	MOV	A, @R1			;  ... get the low byte
	MOV	R7, A			;  ... we keep that in R7
	INC	R1			; and then the high byte
	MOV	A, @R1			;  ...
	MOV	R6, A			;  ... (R6/R7) = (R0)->pbBuffer;
	INC	R1			; cbBuffer is the next field in the BCB
	MOV	A, @R1			;  ... low byte
	MOV	R5, A			;  ...
	INC	R1			;  ...
	MOV	A, @R1			;  ... and high byte 
	MOV	R4, A			;  ... (R4/R5) = (R0)->cbBuffer;

;   Here if the current buffer and is valid and, by implication, so are R4/R5
; and R6/R7....
ISR20:	MOV  	DPL, R7			; load DPTR with the address of the next byte
	MOV  	DPH, R6			;  ...
	MOVX	A, @DPTR		; load the next byte of MP3 data
	INC	DPTR			; and increment the buffer pointer
	MOV	R7, DPL			; save the new address for next time
	MOV	R6, DPH			;  ...

	.if	FAST_SDI
;   In FAST_SDI mode we need to reverse the order of the bits in this byte
; (using m_abReverseBits, of course) and then send it to SBUF.  The 8051
; UART hardware will handle the rest.  Note that it's not necessary to
; check the TI flag here because the 8051 will shift out the byte in only
; 9 machine cycles, and this loop takes far longer than that!
	MOV  	DPTR, #m_abReverseBits	; point to m_abReverseBits
	MOVC 	A, @A+DPTR		; lookup m_abReverseBits[data]
	MOV  	SBUF, A			; and send that to the STA013
	.else
;   In DEBUG (non-FAST_SDI) mode, we have to shift out the eight bits
; one at a time, MSB first, and toggle SCK (serial clock) for each one.
	MOV	R1, #8			; shift out 8 bits
ISR21:	RLC	A			; move the MSB into the carry
	MOV	STA013_ALT_SDI, C	; output the data bit
	SETB	STA013_ALT_SCK		; pulse the clock
	CLR	STA013_ALT_SCK		; ...
	DJNZ	R1, ISR21		; and loop for 8 bits
	.endif

;   Decrement the count of bytes remaining (kept in R4/R5) and, if it's zero,
; put this buffer on the free list and clear R0.  Clearing R0 will make the
; code at the beginning of this ISR attempt to pull a new buffer off the play
; list the next time around.
	DEC	R5			; decrement the LSB of cbBuffer
	CJNE	R5, #0xFF, ISR30	; jump if there's no borrow
	DEC	R4			;  ... and decrement the MSB if there is
ISR30:	MOV	A, R5			; see if cbBuffer == 0
	ORL	A, R4			;  ...
	JZ	ISR31			; jump if this buffer is empty
;   The current buffer still has data.  If the STA013 is still asserting DRQ,
; then loop back and send it another byte.  If, however, the STA013 has rel-
; eased DRQ then we can safely exit from this interrupt.  Remember that the
; DRQ input is active LOW!!!
	JNB	STA013_DATAREQ, ISR20	; jump if the STA013 wants more data
	SJMP	ISR99			; nope - go dismiss this interrupt

;   Here if we've sent all the bytes in this buffer.  First, put the current
; buffer on the free list (otherwise it'll be lost forever!) and the get a
; new buffer...
ISR31:  MOV  	@R0, _g_pFreeBufferList	; (R0)->pNext = m_pFreeBufferList
	MOV  	_g_pFreeBufferList, R0	; m_pFreeBufferList = R0
	MOV	R0, #0			; R0 == NULL indicates no current buffer
;   If the STA013 is still asserting DRQ (remember that its active low!) then
; try to get another buffer now.  Otherwise, we can safely exit now as long
; as R0 contains NULL...
	JNB	STA013_DATAREQ, ISR10	; jump if the STA013 wants more data
;	SJMP	ISR99			; nope - go dismiss this interrupt
	
;   One way or another, we're all done with this interrupt.  Restore the
; original registers and register bank and return...
ISR99:	CLR  	LED_BIT			; (LED off for timing purposes)
	POP  	PSW			; restore register bank and flags
	POP  	DPL			; restore other registers
	POP  	DPH			;  ...
	POP  	ACC			;  ...
	RETI 				; and dismiss the interrupt


;++
; BOOL IsPlayListEmpty (void)
;
;   This routine will return TRUE if the current MP3 buffer queue is empty
; and FALSE if data still remains.  You might be tempted to simply say
; "if (g_pPlayListHead == NULL)", but that doesn't work because MP3DRQ actually
; removes the current buffer from the play list.  Therefore g_pPlayListHead
; will become NULL even though there's still one more buffer left in the
; works...
;--
_IsPlayListEmpty:
;   The first thing is just to test the play list queue - if g_pPlayListHead
; isn't null, then it's guaranteed that the play list isn't empty...
	MOV	DPL, #0			; assume the result is FALSE
	MOV  	A, _g_pPlayListHead	; check the play list queue
	JZ	EMPTY1			; jump if it's empty
	RET				; not empty - return FALSE now

;   Here if the queue is empty - check to see if there's still a buffer "in
; progress" (the pointer to the BCB will in R0 of register bank #1).  Note
; that this code cheats a bit, bit it's must faster than explicitly changing
; register banks (which requires disabling interrupts and the whole thing!).
EMPTY1:	MOV	A, AR10			; fetch R0 from register bank 1
	JZ	EMPTY2			; jump if pBCB == NULL
	RET				; not null - return FALSE

; Here if the play list really is empty...
EMPTY2:	MOV	DPL, #0xFF		; return TRUE
	RET				; and we're outta here
	

;++
; void FlushPlayList (void)
;
;   This routine will flush the MP3 "to be played" buffer queue, plus any
; buffer that's currently playing now.  This doesn't actually stop the
; STA013 decoder, and it's probably a good idea for the caller to do that
; first!
;--
_FlushPlayList:
	PUSH	IE			; all interrupts off while we change
	CLR	EA			;  ... the register bank
	MOV	PSW, #0x08		; select register bank 1

; Put the currently playing buffer (pBCB is in R0) on the free list...
	CJNE	R0, #0, FLUSH1		; is there a current buffer ?
	SJMP	FLUSH2			; nope - skip this
FLUSH1:	MOV  	@R0, _g_pFreeBufferList	; (R0)->pNext = m_pFreeBufferList
	MOV  	_g_pFreeBufferList, R0	; m_pFreeBufferList = R0

; Now return all the buffers in the playlist to the free pool...
FLUSH2:	MOV  	R0, _g_pPlayListHead	; remove the next BCB from the play list
	CJNE	R0, #0, FLUSH3		; jump if we really got a buffer
	SJMP	FLUSH4			; no more buffers on the list
FLUSH3:	MOV  	_g_pPlayListHead, @R0	;  ... m_pPlayListHead = m_pPlayListHead->pNext
	MOV  	@R0, _g_pFreeBufferList	; (R0)->pNext = m_pFreeBufferList
	MOV  	_g_pFreeBufferList, R0	; m_pFreeBufferList = R0
	SJMP	FLUSH2			; and keep flushing

; Reset all the registers and then we can return...
FLUSH4:	SJMP	INIT10			; (re)initialize all registers and return


;++
; void InitializeMP3DRQ (void)
;
;   This routine will initialize the MP3 data request handler.  It will
; configure the 8051's external interrupt hardware as required and it will
; initialize the register bank used by the ISR.  It does NOT, however, turn
; on the interrupts!
;--
_InitializeMP3DRQ:

; Initialize the register bank used by the ISR...
	PUSH	IE		; no interrupts while we've changed
	CLR	EA		;  ... the register bank
	MOV  	PSW, #0x08	; select register bank 1
INIT10:	MOV	A, #0		; clear registers 0..7
	MOV	R0, A		;  ...
	MOV	R1, A		;  ...
	MOV	R2, A		;  ...
	MOV	R3, A		;  ...
	MOV	R4, A		;  ...
	MOV	R5, A		;  ...
	MOV	R6, A		;  ...
	MOV	R7, A		;  ...

;   Initialize the playback buffer queue to empty.  Note that if there
; are any _real_ buffers on this list (if, for example, we just aborted
; a playback) then the caller had better free them before calling this
; routine again!
	MOV	_g_pPlayListHead, #0
	MOV	_g_pPlayListEnd, #0

; Now we can re-enable interrupts...
	MOV	PSW, #0x00	; back to the default register bank
	POP	IE		; and restore interrupts

	.if	FAST_SDI
;   For FAST_SDI mode, configure the 8051's hardware UART to mode 0
; (synchronous shift register).  This mode is automatically clocked
; at 1/6th the CPU clock (on a 6 clock per cycle Philips part; it's
; 1/12th the CPU clock on a regular 8051), so no timer is needed to
; generate a baud rate.
	MOV	SCON, #0x1C	; mode 0
	.else
;   For DEBUG (slow) mode, configure the alternate STA013 clock and
; data outputs....
	CLR	STA013_ALT_SCK	; clock low
	CLR	STA013_ALT_SDI	; and data low too
	.endif

; Configure the 8051's interrupt system...
	CLR	IT0		; INT0 level sensitive
; Set interrupt priority here???
	
	RET
